package net.knarcraft.dynmapcitizens.trait;

import me.blackvein.quests.QuestsAPI;
import me.blackvein.quests.quests.IQuest;
import me.blackvein.quests.quests.IStage;
import net.citizensnpcs.api.CitizensAPI;
import net.citizensnpcs.api.npc.NPC;
import net.citizensnpcs.api.npc.NPCRegistry;
import net.knarcraft.dynmapcitizens.DynmapCitizens;
import net.knarcraft.dynmapcitizens.Icon;
import net.knarcraft.dynmapcitizens.UpdateRate;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.inventory.ItemStack;
import org.dynmap.DynmapAPI;
import org.dynmap.markers.CircleMarker;
import org.dynmap.markers.GenericMarker;
import org.dynmap.markers.MarkerIcon;
import org.dynmap.markers.MarkerSet;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.logging.Level;

/**
 * A handler class for the quests trait
 */
public class QuestsHandler extends AbstractTraitHandler {

    private QuestsAPI questsAPI;
    private MarkerSet questMarkerSet;
    private MarkerSet questAreaMarkerSet;
    private Map<Icon, MarkerIcon> markerIcons;
    private Collection<IQuest> loadedQuests;

    @Override
    public void initialize() {
        questsAPI = (QuestsAPI) Bukkit.getServer().getPluginManager().getPlugin("Quests");
        DynmapAPI dynmapAPI = DynmapCitizens.getInstance().getDynmapAPI();
        markerIcons = DynmapCitizens.getInstance().getMarkerIcons();
        if (questsAPI != null) {
            questMarkerSet = getMarkerSet(dynmapAPI, "quests", "Quests");
            questAreaMarkerSet = getMarkerSet(dynmapAPI, "quest_areas", "Quest areas");
            if (questMarkerSet != null && questAreaMarkerSet != null) {
                questMarkerSet.setHideByDefault(false);
                questAreaMarkerSet.setHideByDefault(true);
                questMarkerSet.setLayerPriority(3);
                questAreaMarkerSet.setLayerPriority(2);
                isEnabled = true;
                return;
            }
        }
        isEnabled = false;
    }

    @Override
    public UpdateRate getUpdateRate() {
        return UpdateRate.VERY_SLOW;
    }

    @Override
    public void updateMarkers() {
        if (questsAPI.isLoading()) {
            return;
        }

        //There is no point in updating if there has been no changes in quests
        boolean questsChanged = loadedQuests == null || !loadedQuests.equals(questsAPI.getLoadedQuests());
        loadedQuests = questsAPI.getLoadedQuests();

        //Remove old quest markers
        questMarkerSet.getMarkers().forEach(GenericMarker::deleteMarker);

        //Updates all quest area markers
        if (questsChanged) {
            updateQuestAreas();
        }

        for (IQuest quest : questsAPI.getLoadedQuests()) {
            UUID npcStartId = quest.getNpcStart();
            //TODO: Store locations of NPCs and see if they've moved?
            // there should probably be a priority and a combination for NPC markers.
            // Use the npc ID as the marker id. If it already exists, try and add additional information. Use priority
            // to decide if icon needs to be changed.
            if (npcStartId != null) {
                String markerDescription = "<b>Quest name:</b> " + quest.getName() + "<br><b>Quest description:</b> " +
                        quest.getDescription() + getQuestStagesInfo(quest);
                addNPCMarker(npcStartId, "Quest Start NPC: ", markerDescription,
                        markerIcons.get(Icon.QUEST_GIVER), questMarkerSet);
            }
            //TODO: Perhaps display requirements and rewards as part of the description

            //Mark NPCs related to this quest
            for (IStage stage : quest.getStages()) {
                markNPCsInQuest(quest, stage, npcStartId);
            }
        }
    }

    /**
     * Updates all quest area markers
     */
    private void updateQuestAreas() {
        questAreaMarkerSet.getCircleMarkers().forEach(GenericMarker::deleteMarker);
        for (IQuest quest : questsAPI.getLoadedQuests()) {
            for (IStage stage : quest.getStages()) {
                markKillLocations(quest, stage);
                markReachLocations(quest, stage);
            }
        }
        //TODO: Mark WorldGuard areas part of quests. Requires WorldGuard integration
        //TODO: See if there is anything to do against overlapping markers
    }

    /**
     * Gets information about a quest's stages
     *
     * @param quest <p>The quest to get information about</p>
     * @return <p>A string with information about the quest's stages</p>
     */
    private String getQuestStagesInfo(IQuest quest) {
        StringBuilder questInfo = new StringBuilder();
        NPCRegistry registry = CitizensAPI.getNPCRegistry();
        for (IStage stage : quest.getStages()) {
            questInfo.append("<br><b>Tasks:</b> ");
            int mobTypes = stage.getMobsToKill().size();
            for (int i = 0; i < mobTypes; i++) {
                questInfo.append("<br>Kill ").append(normalizeName(stage.getMobsToKill().get(i).name())).append(
                        " X ").append(stage.getMobNumToKill().get(i));
            }
            int deliveries = stage.getItemDeliveryTargets().size();
            for (int i = 0; i < deliveries; i++) {
                NPC npc = registry.getByUniqueId(stage.getItemDeliveryTargets().get(i));
                questInfo.append("<br>Deliver ").append(getItemStackString(stage.getItemsToDeliver().get(i))).append(
                        " to ").append(npc.getName());
            }
            if (stage.getFishToCatch() != null) {
                questInfo.append("<br>Catch ").append(stage.getFishToCatch()).append(" fish");
            }
            for (UUID npcId : stage.getNpcsToKill()) {
                questInfo.append("<br>Kill NPC ").append(registry.getByUniqueId(npcId).getName());
            }

            questInfo.append(getQuestItemsTaskString(stage.getBlocksToBreak(), "<br>Break "));
            questInfo.append(getQuestItemsTaskString(stage.getBlocksToCut(), "<br>Cut "));
            questInfo.append(getQuestItemsTaskString(stage.getBlocksToDamage(), "<br>Damage "));
            questInfo.append(getQuestItemsTaskString(stage.getBlocksToUse(), "<br>Use "));
            questInfo.append(getQuestItemsTaskString(stage.getBlocksToPlace(), "<br>Place "));

            questInfo.append(getQuestItemsTaskString(stage.getItemsToBrew(), "<br>Brew "));
            questInfo.append(getQuestItemsTaskString(stage.getItemsToConsume(), "<br>Consume "));
            questInfo.append(getQuestItemsTaskString(stage.getItemsToCraft(), "<br>Craft "));
            questInfo.append(getQuestItemsTaskString(stage.getItemsToEnchant(), "<br>Enchant "));
            questInfo.append(getQuestItemsTaskString(stage.getItemsToSmelt(), "<br>Smelt "));

            int sheepTypes = stage.getSheepToShear().size();
            for (int i = 0; i < sheepTypes; i++) {
                questInfo.append("<br>Shear ").append(stage.getSheepNumToShear().get(i)).append(" ").append(
                        normalizeName(stage.getSheepToShear().get(i).name())).append(" sheep");
            }
            if (stage.getCowsToMilk() != null) {
                questInfo.append("<br>Milk ").append(stage.getCowsToMilk()).append(" cows");
            }

            int mobTamingEntries = stage.getMobsToTame().size();
            for (int i = 0; i < mobTamingEntries; i++) {
                questInfo.append("<br>Tame ").append(stage.getMobNumToTame().get(i)).append(" ").append(
                        normalizeName(stage.getMobsToTame().get(i).name()));
            }
        }
        return questInfo.toString();
    }

    /**
     * Gets a string to display a quest task involving some action on an item
     *
     * @param items       <p>The items that are part of the task</p>
     * @param explanation <p>The explanation of what the player needs to do with the items</p>
     * @return <p>A string describing the necessary tasks</p>
     */
    private String getQuestItemsTaskString(List<ItemStack> items, String explanation) {
        StringBuilder questInfo = new StringBuilder();
        for (ItemStack itemStack : items) {
            questInfo.append(explanation).append(getItemStackString(itemStack));
        }
        return questInfo.toString();
    }

    /**
     * Gets the proper string representation of an item stack
     *
     * @param itemStack <p>The item stack to print</p>
     * @return <p>The string representation of the item stack</p>
     */
    private String getItemStackString(ItemStack itemStack) {
        return normalizeName(itemStack.getType().name()) + " X " + itemStack.getAmount();
    }

    /**
     * Normalizes an internal name to make it human-readable
     *
     * @param name <p>The name to normalize</p>
     * @return <p>The normalized name</p>
     */
    private String normalizeName(String name) {
        return name.toLowerCase().replace("_", " ");
    }

    /**
     * Marks any reach locations found in the given stage
     *
     * @param quest <p>The quest the stage belongs to</p>
     * @param stage <p>The stage to search for reach locations</p>
     */
    private void markReachLocations(IQuest quest, IStage stage) {
        markLocations(stage.getLocationsToReach(), stage.getRadiiToReachWithin(),
                "<b>Target location for:</b> " + quest.getName());
    }

    /**
     * Marks any kill locations found in the given stage
     *
     * @param quest <p>The quest the stage belongs to</p>
     * @param stage <p>The stage to search for kill locations</p>
     */
    private void markKillLocations(IQuest quest, IStage stage) {
        markLocations(stage.getLocationsToKillWithin(), stage.getRadiiToKillWithin(),
                "<b>Kill location for:</b> " + quest.getName());
    }

    /**
     * Marks the given locations on the dynamic map
     *
     * @param locations   <p>The locations to mark</p>
     * @param radii       <p>The radius of each location's circle</p>
     * @param description <p>The description for what the location means</p>
     */
    private void markLocations(List<Location> locations, List<Integer> radii, String description) {
        for (int i = 0; i < locations.size(); i++) {
            Location location = locations.get(i);
            int radius = radii.get(i);

            //Skip if location is invalid
            World world = location.getWorld();
            if (world == null) {
                continue;
            }

            CircleMarker circleMarker = questAreaMarkerSet.createCircleMarker(null, description, true,
                    world.getName(), location.getX(), location.getY(), location.getZ(), radius, radius, false);
            if (circleMarker == null) {
                DynmapCitizens.getInstance().getLogger().log(Level.WARNING, "Unable to create circle marker at " +
                        location + " with radius " + radius);
            } else {
                circleMarker.setFillStyle(0.3, 0x75AFD2);
                circleMarker.setLineStyle(1, 1.0, 0x36c90e);
            }
        }
    }

    /**
     * Marks all NPCs which has interactions as part of this stage
     *
     * @param quest      <p>The quest to mark NPCs for</p>
     * @param stage      <p>The quest stage to mark NPCs for</p>
     * @param npcStartId <p>The id of the quest's start NPC</p>
     */
    private void markNPCsInQuest(IQuest quest, IStage stage, UUID npcStartId) {
        String markerDescription = "Part of quest: " + quest.getName();
        //Mark all interaction targets
        for (UUID interactNpcId : stage.getNpcsToInteract()) {
            addNPCMarker(interactNpcId, "Quest Interact NPC: ", markerDescription,
                    markerIcons.get(Icon.QUEST_INTERACT), questMarkerSet);
        }
        //Mark all kill targets
        for (UUID killNPCId : stage.getNpcsToKill()) {
            addNPCMarker(killNPCId, "Quest Kill NPC: ", markerDescription, markerIcons.get(Icon.QUEST_KILL),
                    questMarkerSet);
        }
        //Mark all delivery targets
        for (UUID deliveryTargetNPCId : stage.getItemDeliveryTargets()) {
            if (!deliveryTargetNPCId.equals(npcStartId)) {
                addNPCMarker(deliveryTargetNPCId, "Quest Delivery NPC: ", markerDescription,
                        markerIcons.get(Icon.QUEST_DELIVER), questMarkerSet);
            }
        }
    }

}
